% (c)2011, Thomas Galliker
% Please report updates/corrections to
% thomas_galliker@bluewin.ch
% 
% Thank you

-module(node2).
-export([start/1,start/2]).
-define(STABILIZE, 1000).
-define(TIMEOUT, 1000).

start(Id) ->
    start(Id, nil).
start(Id, Peer) ->
    timer:start(),
    spawn(fun() -> init(Id, Peer) end).

init(Id, Peer) ->
    Predecessor = nil,
    {ok, Successor} = connect(Id, Peer),
    schedule_stabilize(),
    node(Id, Predecessor, Successor, storage:create()).

connect(Id, nil) ->
    {ok, {Id, self()}};
connect(Id, Peer) ->
    Qref = make_ref(),
    Peer ! {key, Qref, self()},
    receive
	{Qref, Skey} ->
	    % Take the connected node as successor
	    {ok, {Skey, Peer}}
    after ?TIMEOUT ->
	    %io:format("~w: Request timed out.~n",[Id]),
	    % Take ourselfs as successor
	    {ok, {Id, self()}}
    end.

node(Id, Predecessor, Successor, Store) ->
    receive
	{key, Qref, Peer} ->
	    %io:format("~w: Key request from ~w~n", [Id, Peer]),
	    Peer ! {Qref, Id},
	    node(Id, Predecessor, Successor, Store);
	{notify, New} ->
	    %io:format("~w: NOTIFY~n", [Id]),
	    {NewPredecessor, NewStore} = notify(New, Id, Predecessor, Store),
	    %io:format("NewPredecessor=~w: ~n", [NewPredecessor]),
	    %io:format("NewStore=~w: ~n", [NewStore]),
	    node(Id, NewPredecessor, Successor, NewStore);
	{request, Peer} ->
	    %io:format("~w: REQUEST~n", [Id]),
	    request(Peer, Predecessor),
	    node(Id, Predecessor, Successor, Store);
	{status, Pred} ->
	    %io:format("~w: STATUS~n", [Id]),
	    NewSuccessor = stabilize(Pred, Id, Successor),
	    node(Id, Predecessor, NewSuccessor, Store);
	stabilize ->
	    %io:format("~w: STABILIZE~n", [Id]),
	    stabilize(Successor),
	    node(Id, Predecessor, Successor, Store);
	info ->
	    {Pkey, Ppid} = Predecessor,
	    {Skey, Spid} = Successor,
	    io:format("Predecessor ID: ~w // Process ID: ~w~n"
		      "Local Node ID:  ~w // Process ID: ~w~n"
		      "Successor ID:   ~w // Process ID: ~w~n",
		      [Pkey, Ppid, Id, self(), Skey, Spid]),
	    node(Id, Predecessor, Successor, Store);
	ring ->
	    % Start ring broadcast
	    {_, Spid} = Successor,
	    Spid ! {ring, {Id, []}},
	    node(Id, Predecessor, Successor, Store);
	{ring, {Id, Info}} ->
	    % Show ring broadcast result
	    io:format("~w: Ring broadcast finished:~n"
		      "~w~n", [Id, Info]),
	    node(Id, Predecessor, Successor, Store);
	{ring, {CallerId, Info}} ->

	    % Pass ring broadcast further
	    io:format("~w: Ring broadcast passed by~n"
		      "~w~n", [Id, Info]),
	    {_, Spid} = Successor,
	    Spid ! {ring, {CallerId, [Id|Info]}},
	    node(Id, Predecessor, Successor, Store);
	probe ->
	    create_probe(Id, Successor),
	    node(Id, Predecessor, Successor, Store);
	{probe, Id, Nodes, T} ->
	    % If the Probe is equal to the Id of the node
	    % we know that we sent it and can report
	    % the time it took to pass it around the ring.
	    remove_probe(T, Nodes),
	    node(Id, Predecessor, Successor, Store);
	{probe, Ref, Nodes, T} ->
	    forward_probe(Id, Ref, T, Nodes, Successor),
	    node(Id, Predecessor, Successor, Store);

	% Key store related commands
	{add, Key, Value, Qref, Client} ->
	    AddedStore = add(Key, Value, Qref, Client,
			Id, Predecessor, Successor, Store),
	    node(Id, Predecessor, Successor, AddedStore);
	{lookup, Key, Qref, Client} ->
	    lookup(Key, Qref, Client, Id, Predecessor, Successor, Store),
	    node(Id, Predecessor, Successor, Store);
	{handover, Elements} ->
	    MergedStore = storage:merge(Store, Elements),
	    node(Id, Predecessor, Successor, MergedStore);
	print ->
	    io:format("~w: Store entries:~n"
		      "~p~n", [Id, Store]),
	    node(Id, Predecessor, Successor, Store);

	UnknownMessage ->
	    io:format("~w: Error: Unknown message received: ~w~n", [Id, UnknownMessage]),
	    node(Id, Predecessor, Successor, Store)
	end.

schedule_stabilize() ->
    timer:send_interval(?STABILIZE, self(), stabilize).

stabilize({_, Spid}) ->
    Spid ! {request, self()}.

stabilize(Pred, Id, Successor) ->
    {Skey, Spid} = Successor,
    case Pred of
	nil ->
	    % We do not know about a predecessor
	    Spid ! {notify, {Id, self()}},
	    Successor;
	{Id, _} ->
	    % We are pointing to ourselfs.
	    Successor;
	{Skey, _} ->
	    % Our successor uses a predecessor which points to itself.
	    % Let's notify him, that he can point to us.
	    Spid ! {notify, {Id, self()}},
	    Successor;
	{Xkey, Xpid} ->
	    case key:between(Xkey, Id, Skey) of
		true ->
		    % If the key of the predecessor of our successor (Xkey)
		    % is between us and our successor we should adopt
		    % this node as our successor and run stabilization again.
		    stabilize(Pred, Id, {Xkey, Xpid});	    
		false ->
		    % If we should be in between the nodes
		    % we inform our successor of our existence.
		    Spid ! {notify, {Id, self()}},			    
		    Successor
		end
    end.

request(Peer, Predecessor) ->
    case Predecessor of
	nil ->
	    Peer ! {status, nil};
	{Pkey, Ppid} ->
	    Peer ! {status, {Pkey, Ppid}}
    end.

notify({Nkey, Npid}, Id, Predecessor, Store) ->
    case Predecessor of
	nil ->
	    % If our own predecessor is set to nil
	    % we take the received predecessor
	    Keep = handover(Store, Nkey, Npid),
	    {{Nkey, Npid}, Keep};
	{Pkey, _} ->
	    case key:between(Nkey, Pkey, Id) of
		true ->
		    % Nkey is the better (closer) predecessor
		    % than Pkey! Let's take the new predecessor
		    Keep = handover(Store, Nkey, Npid),
		    {{Nkey, Npid}, Keep};
		false ->
		    {Predecessor, Store}
	    end
    end.

handover(Store, Nkey, Npid) ->
    {Keep, Leave} = storage:split(Nkey, Store),
    Npid ! {handover, Leave},
    Keep.

create_probe(Id, {_, Spid}) ->
	Spid ! {probe, Id, [Id], now()}.

remove_probe(T, Nodes) ->
	{_, BeforeS, BeforeNs} = T,
	{_, NowS, NowNs} = now(),
	Time = (NowS-BeforeS) * 1000000 + (NowNs-BeforeNs),
	io:format("Probe successfully finished:~n"
		  "Round trip duration: ~w ns~n"
		  "Ring nodes: ~w~n",[Time, Nodes]).

forward_probe(Id, Ref, T, Nodes, {_, Spid}) ->
	Spid ! {probe, Ref, [Id|Nodes], T}.

% To add a new key value we must first determine
% if our node is the node that should take care of the key.
add(Key, Value, Qref, Client, Id, {Pkey, _}, {_, Spid}, Store) ->
    % A node will take care of all keys from (but not including) the identifier
    % of its predecessor to (and including) the identifier of itself.
    % If we are not responsible we simply send a add message to our successor.
    case key:between(Key, Pkey, Id) of
	true ->
	    Client ! {Qref, ok},
            % Add key to the store
	    storage:add(Key, Value, Store);
	false ->
            % If we are not responsible we simply
	    % send a add message to our successor.
	    Spid ! {add, Key, Value, Qref, Client},
	    Store
    end.


lookup(Key, Qref, Client, Id, {Pkey, _}, Successor, Store) ->
    case key:between(Key, Pkey, Id) of
	true ->
	    Result = storage:lookup(Key, Store),
	    Client ! {Qref, Result};
	false ->
	    {_, Spid} = Successor,
	    Spid ! {lookup, Key, Qref, Client}    
    end.

