% (c)2011, Thomas Galliker
% Please report updates/corrections to
% thomas_galliker@bluewin.ch
% 
% Thank you

-module(node3).
-export([start/1,start/2]).
-compile({no_auto_import,[demonitor/1]}).
-define(STABILIZE, 100).
-define(TIMEOUT, 1000).

start(Id) ->
    start(Id, nil).
start(Id, Peer) ->
    timer:start(),
    spawn(fun() -> init(Id, Peer) end).

init(Id, Peer) ->
    Predecessor = nil,
    Store = storage:create(),
    {ok, Successor} = connect(Id, Peer),
    schedule_stabilize(),
    node(Id, Predecessor, Successor, nil, Store).

connect(Id, nil) ->
    {ok, {Id, monitor(self()), self()}};
connect(Id, Peer) ->
    Qref = make_ref(),
    Peer ! {key, Qref, self()},
    receive
	{Qref, Skey} ->
	    % Take the connected node as successor
	    {ok, {Skey, monitor(Peer), Peer}}
    after ?TIMEOUT ->
	    %io:format("~w: Request timed out.~n",[Id]),
	    % Take ourselfs as successor
	    {ok, {Id, monitor(self()), self()}}
    end.

node(Id, Predecessor, Successor, Next, Store) ->
    receive
	{key, Qref, Peer} ->
	    %io:format("~w: Key request from ~w~n", [Id, Peer]),
	    Peer ! {Qref, Id},
	    node(Id, Predecessor, Successor, Next, Store);
	{notify, New} ->
	    %io:format("~w: Notification from node ~w~n", [Id, New]),
	    {NewPredecessor, NewStore} = notify(New, Id, Predecessor, Store),
	    %io:format("NewPredecessor=~w: ~n", [NewPredecessor]),
	    %io:format("NewStore=~w: ~n", [NewStore]),
	    node(Id, NewPredecessor, Successor, Next, NewStore);
	{request, Peer} ->
	    %io:format("~w: REQUEST~n", [Id]),
	    request(Peer, Predecessor, Successor),
	    node(Id, Predecessor, Successor, Next, Store);
	{status, Pred, Nx} ->
	    %io:format("~w: STATUS~n", [Id]),
	    {NewSuccessor, NewNx} = stabilize(Pred, Nx, Id, Successor),
	    node(Id, Predecessor, NewSuccessor, NewNx, Store);
	stabilize ->
	    %io:format("~w: STABILIZE~n", [Id]),
	    stabilize(Successor),
	    node(Id, Predecessor, Successor, Next, Store);
	{'DOWN', Ref, process, _, _} ->
	    io:format("~w: Received DOWN~n", [Id]),
	    {Pred, Succ, Nxt} = down(Ref, Predecessor, Successor, Next),
	    node(Id, Pred, Succ, Nxt, Store);

	info ->
	    {Pkey, _, Ppid} = Predecessor,
	    {Skey, _, Spid} = Successor,
	    io:format("Predecessor ID: ~w // Process ID: ~w~n"
		      "Local Node ID:  ~w // Process ID: ~w~n"
		      "Successor ID:   ~w // Process ID: ~w~n",
		      [Pkey, Ppid, Id, self(), Skey, Spid]),
	    node(Id, Predecessor, Successor, Next, Store);

	probe ->
	    create_probe(Id, Successor),
	    node(Id, Predecessor, Successor, Next, Store);
	{probe, Id, Nodes, T} ->
	    % If the Probe is equal to the Id of the node
	    % we know that we sent it and can report
	    % the time it took to pass it around the ring.
	    remove_probe(T, Nodes),
	    node(Id, Predecessor, Successor, Next, Store);
	{probe, Ref, Nodes, T} ->
	    forward_probe(Id, Ref, T, Nodes, Successor),
	    node(Id, Predecessor, Successor, Next, Store);

	% Key store related commands
	{add, Key, Value, Qref, Client} ->
	    AddedStore = add(Key, Value, Qref, Client,
			Id, Predecessor, Successor, Store),
	    node(Id, Predecessor, Successor, Next, AddedStore);
	{lookup, Key, Qref, Client} ->
	    lookup(Key, Qref, Client, Id, Predecessor, Successor, Store),
	    node(Id, Predecessor, Successor, Next, Store);
	{handover, Elements} ->
	    MergedStore = storage:merge(Store, Elements),
	    node(Id, Predecessor, Successor, Next, MergedStore);
	print ->
	    io:format("~w: Store entries:~n"
		      "~p~n", [Id, Store]),
	    node(Id, Predecessor, Successor, Next, Store);
	stop ->
	    io:format("~w: Received STOP command...~n", [Id]);
	stopall ->
	    io:format("~w: Received STOPALL command...~n", [Id]),
	    forward_stop(Successor);

	UnknownMessage ->
	    io:format("~w: Error: Unknown message received: ~w~n", [Id, UnknownMessage]),
	    node(Id, Predecessor, Successor, Next, Store)
	end.

schedule_stabilize() ->
    timer:send_interval(?STABILIZE, self(), stabilize).

stabilize({_, _, Spid}) ->
    Spid ! {request, self()}.

stabilize(Pred, Next, Id, Successor) ->
    {Skey, Smon, Spid} = Successor,
    case Pred of
	nil ->
	    % We do not know about a predecessor
	    Spid ! {notify, {Id, self()}},
	    {Successor, Next};
	{Id, _} ->
	    % We are pointing to ourselfs
	    {Successor, Next};
	{Skey, _} ->
	    % Our successor uses a predecessor which points to itself.
	    % Let's notify him, that he can point to us.
	    Spid ! {notify, {Id, self()}},
	    {Successor, Next};
	{Xkey, Xpid} ->
	    case key:between(Xkey, Id, Skey) of
		true ->
		    % If the key of the predecessor of our successor (Xkey)
		    % is between us and our successor we should adopt
		    % this node as our successor and run stabilization again.
		    demonitor(Smon),
		    stabilize(Pred, Next, Id, {Xkey, monitor(Xpid), Xpid});	    
		false ->
		    % If we should be in between the nodes
		    % we inform our successor of our existence.
		    Spid ! {notify, {Id, self()}},
		    {Successor, Next}
		end
    end.

request(Peer, Predecessor, {Skey, _, Spid}) ->
    case Predecessor of
	nil ->
	    Peer ! {status, nil, {Skey, Spid}};
	{Pkey, _, Ppid} ->
	    Peer ! {status, {Pkey, Ppid}, {Skey, Spid}}
    end.

notify({Nkey, Npid}, Id, Predecessor, Store) ->
    case Predecessor of
	nil ->
	    % If our own predecessor is set to nil
	    % we take the received predecessor
	    Keep = handover(Store, Nkey, Npid),
	    {{Nkey, monitor(Npid), Npid}, Keep};
	{Pkey, Pmon, _} ->
	    case key:between(Nkey, Pkey, Id) of
		true ->
		    % Nkey is the better (closer) predecessor
		    % than Pkey! Let's take the new predecessor
		    Keep = handover(Store, Nkey, Npid),
		    demonitor(Pmon),
		    io:format("~w: New predecessor found: ~w ~w~n", [Id,Nkey,Npid]),
		    {{Nkey, monitor(Npid), Npid}, Keep};
		false ->
		    {Predecessor, Store}
	    end
    end.

handover(Store, Nkey, Npid) ->
    {Leave, Keep} = storage:split(Nkey, Store),
    Npid ! {handover, Leave},
    Keep.

create_probe(Id, {_, _, Spid}) ->
	Spid ! {probe, Id, [Id], now()}.

remove_probe(T, Nodes) ->
	{_, BeforeS, BeforeNs} = T,
	{_, NowS, NowNs} = now(),
	Time = (NowS-BeforeS) * 1000000 + (NowNs-BeforeNs),
	io:format("Probe successfully finished:~n"
		  "Round trip duration: ~w ns~n"
		  "Ring nodes: ~w~n",[Time, Nodes]).

forward_probe(Id, Ref, T, Nodes, {_, _, Spid}) ->
	Spid ! {probe, Ref, [Id|Nodes], T}.

forward_stop({_, _, Spid}) ->
	Spid ! stopall.

% To add a new key value we must first determine
% if our node is the node that should take care of the key.
add(Key, Value, Qref, Client, Id, {Pkey, _, _}, {_, _, Spid}, Store) ->
    % A node will take care of all keys from (but not including) the identifier
    % of its predecessor to (and including) the identifier of itself.
    % If we are not responsible we simply send a add message to our successor.
    case key:between(Key, Pkey, Id) of
	true ->
	    Client ! {Qref, ok},
            % Add key to the store
	    storage:add(Key, Value, Store);
	false ->
            % If we are not responsible we simply
	    % send a add message to our successor.
	    Spid ! {add, Key, Value, Qref, Client},
	    Store
    end.


lookup(Key, Qref, Client, Id, {Pkey, _, _}, Successor, Store) ->
    %Use the same distinction as in the add function
    case key:between(Key, Pkey, Id) of
	true ->
	    Result = storage:lookup(Key, Store),
	    Client ! {Qref, Result, Id};
	false ->
	    {_, _, Spid} = Successor,
	    Spid ! {lookup, Key, Qref, Client}    
    end.

monitor(Pid) ->
    erlang:monitor(process, Pid).
demonitor(nil) ->
    ok;
demonitor(Pid) ->
    erlang:demonitor(Pid, [flush]).

down(Ref, {_, Ref, _}, Successor, Next) ->
	{nil, Successor, Next};
down(Ref, Predecessor, {_, Ref, _}, {Nkey, Npid}) ->
	demonitor(Ref),
	Nref = monitor(Npid),
	{Predecessor, {Nkey, Nref, Npid}, nil}.

